/*
 * Copyright (c) 1993 Addison-Wesley Publishing Company.
 * All rights reserved.  This source code is provided
 * for its instructional value, and is not guaranteed
 * for any particular purpose.
 */
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

extern void error(const char *fmt, ...);
int catmap(int);
void catreg(int);

void
main(int argc, char *argv[])
{
	int i, fd;

	/*
	 * Process each file named on the command line.
	 */
	for (i = 1; i < argc; i++) {
		if ((fd = open(argv[i], O_RDONLY)) < 0) {
			error("cat: cannot open %s", argv[i]);
			continue;
		}

		/*
		 * Try mapping the file first.  If that fails,
		 * fall back to using read(2).
		 */
		if (catmap(fd) < 0)
			catreg(fd);
		close(fd);
	}
	exit(0);
}

int
catmap(int fd)
{
	int nw;
	caddr_t addr;
	struct stat sbuf;

	/*
	 * Get the file size and try to map the whole file.
	 */
	if (fstat(fd, &sbuf) < 0)
		return(-1);
	addr = mmap(0, sbuf.st_size, PROT_READ, MAP_PRIVATE,
	  fd, 0);
	if (addr == (caddr_t)-1) /* mmap failed */
		return(-1);

	/*
	 * Write the file just mapped.  Clear errno
	 * to silence error() on short writes.
	 */
	errno = 0;
	nw = write(1, addr, sbuf.st_size);
	if (nw != sbuf.st_size) {
		if (nw < 0)
			error("cat: write failed");
		else
			error("cat: short write");
	}

	/*
	 * Undo the mapping.
	 */
	munmap(addr, sbuf.st_size);
	return(0);
}
